/*********************************************
 * vim:sw=8:ts=8:si:et
 * To use the above modeline in vim you must have "set modeline" in your .vimrc
 * Author: Guido Socher
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Ethernet host watchdog
 *
 * Chip type           : Atmega168 or Atmega328 with ENC28J60
 * Note: there is a version number in the text. Search for tuxgraphics
 *********************************************/
#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include "ip_arp_udp_tcp.h"
#include "enc28j60.h"
#include "timeout.h"
#include "websrv_help_functions.h"

// set output to VCC, red LED off
#define LEDOFF PORTB|=(1<<PORTB1)
// set output to GND, red LED on
#define LEDON PORTB&=~(1<<PORTB1)
// to test the state of the LED
#define LEDISOFF PORTB&(1<<PORTB1)

// please modify the following two lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices:
static uint8_t mymac[6] = {0x54,0x55,0x58,0x11,0x00,0x29};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
static uint8_t myip[4] = {10,0,0,29};
static uint8_t monitoredhost[4] = {10,0,0,7};
// the password string (only a-z,0-9,_ characters):
static char password[9]="secret"; // must not be longer than 8 char
//
static uint8_t gwip[4]; // optional gwip
// listen port for tcp/www:
static uint16_t mywwwport=80;
//------------------- do not change below this line
static volatile uint8_t hoststatus=1; // 0=all ping/replies OK, otherwise amount of missed counts
static uint8_t resetcount=0; // how often reset was executed
static uint8_t hostbackafterreboot=0; // wait for on answer until we really activate the wd
static uint8_t sendping=1; // 1 we send ping (and receive ping), 0 we receive ping only
static volatile uint8_t pingnow=1; // 1 means time has run out send a ping now
static volatile uint8_t resetnow=0; 
static uint8_t reinitmac=0; 
static uint8_t do_extra_arp=0; 
static uint8_t pinginterval=10; // after how many seconds to send or receive a ping (value range: 2 - 250)
static volatile uint8_t pingtimer=1; // > 0 means wd running
//
#define BUFFER_SIZE 680
static uint8_t buf[BUFFER_SIZE+1];
static uint16_t gPlen;
// global string buffer
#define STR_BUFFER_SIZE 24
static char gStrbuf[STR_BUFFER_SIZE+1];
//
// 
uint8_t verify_password(char *str)
{
        // the first characters of the received string are
        // a simple password/cookie:
        if (strncmp(password,str,strlen(password))==0){
                return(1);
        }
        return(0);
}

uint16_t http200ok(void)
{
        return(fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\n\r\n")));
}


// prepare the webpage by writing the data to the tcp send buffer
uint16_t print_webpage(uint8_t *buf)
{
        uint16_t plen;
        // periodic refresh every 60sec:
        //plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\nRefresh: 60\r\n\r\n"));
        // no periodic refresh:
        plen=http200ok();
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/now>[actions]</a> <a href=http://tuxgraphics.org/hwd3.html>[help]</a> <a href=/>[refresh]</a>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Host watchdog</h2>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("Status: "));
        // if watchdog is off then status is unknown:
        if (pingtimer==0){
                plen=fill_tcp_data_p(buf,plen,PSTR("-"));
        }else{
                if (hoststatus==0){
                        plen=fill_tcp_data_p(buf,plen,PSTR("<font color=#00ff00>OK</font>"));
                }else{
                        itoa(hoststatus,gStrbuf,10);
                        plen=fill_tcp_data(buf,plen,gStrbuf);
                }
        }
        plen=fill_tcp_data_p(buf,plen,PSTR(" [reset cnt:"));
        itoa(resetcount,gStrbuf,10);
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR(", state: "));
        if (pingtimer==0){
                plen=fill_tcp_data_p(buf,plen,PSTR("stopped"));
        }else{
                if (hostbackafterreboot){
                        plen=fill_tcp_data_p(buf,plen,PSTR("active"));
                }else{
                        plen=fill_tcp_data_p(buf,plen,PSTR("passive"));
                }
        }
        plen=fill_tcp_data_p(buf,plen,PSTR("]\n<br><hr><pre><form action=/mod method=get>"));
        plen=fill_tcp_data_p(buf,plen,PSTR(" monitored IP: <input type=text name=ip value="));
        mk_net_str(gStrbuf,monitoredhost,4,'.',10);
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR("> GW: <input type=text name=gp value="));
        mk_net_str(gStrbuf,gwip,4,'.',10);
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR(">\nping interval: <input type=text name=pi value="));
        itoa(pinginterval,gStrbuf,10);
        plen=fill_tcp_data(buf,plen,gStrbuf);
        plen=fill_tcp_data_p(buf,plen,PSTR("> <input type=checkbox name=sp value=1 "));
        if (sendping){
                plen=fill_tcp_data_p(buf,plen,PSTR("checked"));
        }
        plen=fill_tcp_data_p(buf,plen,PSTR(">Send ping\n     password: <input type=password name=pw>\n<input type=submit value=\"change\"></form>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("\n<hr>3.0, tuxgraphics"));

        return(plen);
}


uint16_t print_webpage_now(uint8_t *buf)
{
        uint16_t plen;
        plen=http200ok();
        plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/>[home]</a> <a href=http://tuxgraphics.org/hwd3.html>[help]</a>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>Actions</h2>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<form action=/ack method=get>"));
        plen=fill_tcp_data_p(buf,plen,PSTR("Passw: <input type=password size=8 name=pw><input type=hidden name=rst value=1><input type=submit value=\"reboot host now\"></form>\n"));
        plen=fill_tcp_data_p(buf,plen,PSTR("<hr><form action=/ack method=get>"));
        if (pingtimer>0){
                plen=fill_tcp_data_p(buf,plen,PSTR("Passw: <input type=password size=8 name=pw><input type=hidden name=stp value=1><input type=submit value=\"stop watchdog now\"></form><hr>"));
        }else{
                plen=fill_tcp_data_p(buf,plen,PSTR("Passw: <input type=password size=8 name=pw><input type=hidden name=srt value=1><input type=submit value=\"start watchdog now\"></form><hr>"));
        }
        return(plen);
}

uint16_t print_webpage_confirm(uint8_t *buf)
{
        uint16_t plen;
        plen=http200ok();
        plen=fill_tcp_data_p(buf,plen,PSTR("<h2>OK</h2> <a href=/>-&gt;continue</a>\n"));
        return(plen);
}

void resethost(void)
{
        PORTD|= (1<<PIND7);// transistor on
        resetcount++;
        hostbackafterreboot=0;
        resetnow=1; // trigger counter to clear transistor 2 sec later
}

void store_in_eeprom(void)
{
        eeprom_write_byte((uint8_t *)20,19); // magic number
        eeprom_write_block((uint8_t *)gwip,(void *)21,sizeof(gwip));
        eeprom_write_block((uint8_t *)monitoredhost,(void *)26,sizeof(monitoredhost));
        eeprom_write_byte((uint8_t *)31,pinginterval); 
        eeprom_write_byte((uint8_t *)32,sendping); 
        reinitmac=1;
        hostbackafterreboot=0;
        resetcount=0;
        hoststatus=1;
}

// takes a string of the form ack?pw=xxx&rst=1 and analyse it
// the first "/" is already removed
// return values:  0 error: 404 Not Found
//                -1 error: 401 Unauthorized
//                 1 main page
int8_t analyse_get_url(char *str)
{
        static uint8_t tmpip[4];
        if (str[0]==' '){
                return(1); // end of url, main page
        }
        if (strncmp("now",str,3)==0){
                gPlen=print_webpage_now(buf);
                return(10);
        }
        if (strncmp("favicon.ico",str,11)==0){
                gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 301 Moved Permanently\r\nLocation: "));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("http://tuxgraphics.org/ico/wd.ico"));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("\r\n\r\nContent-Type: text/html\r\n\r\n"));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("<h1>301 Moved Permanently</h1>\n"));
                return(10);
        }
        // --------
        // actions: reboot stop start
        if (strncmp("ack",str,3)==0){
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"pw")){
                        urldecode(gStrbuf);
                        if (verify_password(gStrbuf)){
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"rst")){
                                        resethost();
                                        gPlen=print_webpage_confirm(buf);
                                        return(10);
                                }
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"stp")){
                                        pingtimer=0; // wd off 
                                        LEDOFF;
                                        gPlen=print_webpage_confirm(buf);
                                        return(10);
                                }
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"srt")){
                                        pingtimer=1; // wd on 
                                        hoststatus=1;
                                        if (sendping && pinginterval>5){
                                                pingtimer=pinginterval-4;
                                        }
                                        gPlen=print_webpage_confirm(buf);
                                        return(10);
                                }
                        }
                }
                return(-1);
        }
        // modify settings:
        if (strncmp("mod",str,3)==0){
                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"pw")){
                        urldecode(gStrbuf);
                        if (verify_password(gStrbuf)){
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"ip")){
                                        urldecode(gStrbuf);
                                        if (parse_ip(tmpip,gStrbuf)){
                                                return(0); // not a good IP
                                        }
                                        // good ip:
                                        monitoredhost[0]=tmpip[0];
                                        monitoredhost[1]=tmpip[1];
                                        monitoredhost[2]=tmpip[2];
                                        monitoredhost[3]=tmpip[3];
                                }
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"gp")){
                                        urldecode(gStrbuf);
                                        if (parse_ip(tmpip,gStrbuf)){
                                                // not a good IP, do not use GW:
                                                gwip[0]=0;
                                                gwip[1]=0;
                                                gwip[2]=0;
                                                gwip[3]=0;

                                        }
                                        // good ip:
                                        gwip[0]=tmpip[0];
                                        gwip[1]=tmpip[1];
                                        gwip[2]=tmpip[2];
                                        gwip[3]=tmpip[3];
                                }
                                // some default values:
                                pinginterval=5;
                                sendping=0;
                                //
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"sp")){
                                        sendping=1;
                                }
                                if (find_key_val(str,gStrbuf,STR_BUFFER_SIZE,"pi")){
                                        urldecode(gStrbuf);
                                        pinginterval=atoi(gStrbuf);
                                        if (pinginterval<2){
                                                pinginterval=2;
                                        }
                                        if (pinginterval>250){
                                                pinginterval=250;
                                        }
                                        // accellerate the status update:
                                        if (sendping && pinginterval>5){
                                                pingtimer=pinginterval-4;
                                        }
                                }
                                store_in_eeprom();
                                gPlen=print_webpage_confirm(buf);
                                return(10);
                        }
                }
                return(-1);
        }
        return(0);
}

void step_hoststatus(void)
{
        hoststatus++; // get's cleared as soon as we get a ping
        // refresh arp in case mac has changed:
        if (sendping==1 && hoststatus>1){
                do_extra_arp=1;
        }
        if (hoststatus==0xff){
                hoststatus=1; // wrap to 1, not zero
        }
        LEDON;
}

// timer interrupt, step counter every second
ISR(TIMER1_COMPA_vect){
        //-----
        if (resetnow>2){
                PORTD &= ~(1<<PIND7);// transistor off
                resetnow=0;
        }
        if (resetnow){
                resetnow++;
        }
        if (hoststatus>5 && hostbackafterreboot){
                resethost();
        }
        //-----
        if (pingtimer>pinginterval){
                // do not wrap to 0 continue with 1:
                pingtimer=1;
        }
        //----- set a flag to send a ping as soon as possible 
        // this is to avoid interference with web traffic
        if (pingtimer == pinginterval){
                pingnow=1;
        }
        // pingtimer==0 means off
        if (pingtimer>0){
                if (sendping==0 && pingnow){
                        // we should receive a ping
                        pingnow=0;
                        step_hoststatus();
                }
                pingtimer++;
        }
        // for security reasons as we increas pingtimer also below in main
        if (pingtimer>252){
                pingtimer=1;
        }
}


// Generate an interrup about ever 1s form the 12.5MHz system clock
// Since we have that 1024 prescaler we do not really generate a second
// (1.00000256000655361677s) 
void timer_init(void)
{
        /* write high byte first for 16 bit register access: */
        TCNT1H=0;  /* set counter to zero*/
        TCNT1L=0;
        // Mode 4 table 14-4 page 132. CTC mode and top in OCR1A
        // WGM13=0, WGM12=1, WGM11=0, WGM10=0
        TCCR1A=(0<<COM1B1)|(0<<COM1B0)|(0<<WGM11);
        TCCR1B=(1<<CS12)|(1<<CS10)|(1<<WGM12)|(0<<WGM13); // crystal clock/1024

        // At what value to cause interrupt. You can use this for calibration
        // of the clock. Theoretical value for 12.5MHz: 12207=0x2f and 0xaf
        OCR1AH=0x2f;
        OCR1AL=0xaf;
        // interrupt mask bit:
        TIMSK1 = (1 << OCIE1A);
}

void ping_callback(uint8_t *from_ip){
        uint8_t i=0;
        while(i<4){
                if(from_ip[i]!=monitoredhost[i]){
                        // not the host we wanted to be pinged from:
                        return;
                }
                i++;
        }
        hoststatus=0;
        hostbackafterreboot=1;
        LEDOFF;
}

int main(void){

        uint16_t pktlen;
        int8_t cmd;
        
        // set the clock speed to "no pre-scaler" (8MHz with internal osc or 
        // full external speed)
        // set the clock prescaler. First write CLKPCE to enable setting of clock the
        // next four instructions.
        CLKPR=(1<<CLKPCE); // change enable
        CLKPR=0; // "no pre-scaler"
        _delay_loop_1(0); // 60us

        // LED
        DDRB|= (1<<DDB1); // enable PB1, LED as output
        LEDOFF;

        // the transistor on PD7
        DDRD|= (1<<DDD7);
        PORTD &= ~(1<<PIND7);// transistor off

        // read eeprom values unless, there are no valid values.
        if (eeprom_read_byte((uint8_t *)20) == 19 ){
                // ok magic number matches accept values
                eeprom_read_block((uint8_t *)gwip,(void *)21,sizeof(gwip));
                eeprom_read_block((uint8_t *)monitoredhost,(void *)26,sizeof(monitoredhost));
                pinginterval=eeprom_read_byte((uint8_t *)31);
                sendping=eeprom_read_byte((uint8_t *)32);
        }
        // accellerate the status update:
        if (sendping && pinginterval>5){
                pingtimer=pinginterval-4;
        }

        //initialize enc28j60:
        enc28j60Init(mymac);
        enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
        _delay_loop_1(0); // 60us

        /* Magjack leds configuration, see enc28j60 datasheet, page 11 */
        // LEDB=yellow LEDA=green
        //
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);
        _delay_loop_1(0); // 60us

        //init the ethernet/ip layer:
        init_ip_arp_udp_tcp(mymac,myip,mywwwport);
        if (gwip[0]==0){
                client_set_gwip(monitoredhost);
        }else{
                client_set_gwip(gwip);  // e.g internal IP of dsl router
        }
        timer_init();

        sei(); // interrupt enable

        // this is where we check if we got a ping from the monitored host:
        register_ping_rec_callback(&ping_callback);

        while(1){
                pktlen=enc28j60PacketReceive(BUFFER_SIZE, buf);
                gPlen=packetloop_icmp_tcp(buf,pktlen);
                if (pktlen){
                        // get the next new packet:
                        if(packetloop_icmp_checkreply(buf,monitoredhost)){
                                // ping reply is from monitored host and ping was from us
                                hoststatus=0;
                                hostbackafterreboot=1;
                                LEDOFF;
                                continue;
                        }
                }else{
                        // pktlen==0, nothing to do from network side
                        //
                        if (do_extra_arp){
                                do_extra_arp=0;
                                if (gwip[0]==0){
                                        client_arp_whohas(buf,monitoredhost);
                                }else{
                                        client_arp_whohas(buf,gwip);  
                                }
                        }
                        if (reinitmac){
                                reinitmac=0;
                                if (gwip[0]==0){
                                        client_set_gwip(monitoredhost);
                                }else{
                                        client_set_gwip(gwip);  // e.g internal IP of dsl router
                                }
                        }
                }
                // spontanious messages must not interfer with
                // web pages. We are idle here:
                if(gPlen==0){
                        if (sendping && client_waiting_gw()==0 && pingnow){
                                pingnow=0;
                                client_icmp_request(buf,monitoredhost);
                                pingtimer++; // otherweise we would call this function again
                                step_hoststatus();
                        }
                        continue;
                }

                if (strncmp("GET ",(char *)&(buf[gPlen]),4)!=0){
                        // head, post and other methods:
                        gPlen=http200ok();
                        gPlen=fill_tcp_data_p(buf,gPlen,PSTR("<h1>200 OK</h1>"));
                        goto SENDTCP;
                }
                cmd=analyse_get_url((char *)&(buf[gPlen+5]));
                // for possible status codes see:
                // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                if (cmd==-1){
                        gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 401 Unauthorized\r\nContent-Type: text/html\r\n\r\n<h1>401 Unauthorized</h1><a href=/>continue</a>\n"));
                        goto SENDTCP;
                }
                if (cmd==0){
                        gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 404 Not Found\r\nContent-Type: text/html\r\n\r\n<h1>404 Not Found</h1>"));
                        goto SENDTCP;
                }
                if (cmd==10){
                        goto SENDTCP;
                }
                // cmd==1 goes here
                gPlen=print_webpage(buf);
                //
SENDTCP:
                www_server_reply(buf,gPlen); // send data
        }
        return (0);
}
